#include <vdmgr.h>
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>
#include "lua.h"


typedef struct _vdmodlua			VDMODLUA ;


	enum {
		CLOCKEVENT_MAX				=	8,							//!< NbNCxgő吔
		PIN_MAX						=	32,							//!< ső吔
		MEMORY_NAME_SIZE			=	64,							//!< [őTCY
		REGISTER_MAX				=	64,							//!< WX^ő吔
		REGISTER_NAME_SIZE			=	512							//!< WX^őTCY
	} ;

	//! NbNCxg
	typedef struct {
		VDMODCLOCKEVENT				clockEvent ;					//!< NbNCxg
		VDMODLUA					*p ;							//!< Lua W[
	} CLOCKEVENTINFO ;

	//! s
	typedef struct {
		VDLIBPIN					pin ;							//!< s
		VDMODLUA					*p ;							//!< Lua W[
	} PININFO ;

	//! Lua
	struct _vdmodlua {
		VDMGR						*p_vdmgr ;						//!< zfoCXǗ
		IVDMOD						vdmod ;							//!< W[C^tF[X
		CLOCKEVENTINFO				clockEventInfo[CLOCKEVENT_MAX] ;//!< NbNCxg
		IVDMODCLOCKGEN				*p_clockGen ;					//!< NbNWFl[^C^tF[X
		PININFO						pinInfo[PIN_MAX] ;				//!< s
		IVDMODADDRESSBUS			addressBus ;					//!< AhXoXC^tF[X
		IVDMODADDRESSBUS			*p_addressBus ;					//!< AhXoXC^tF[X
		IVDMODMEMORYCTRL			memoryCtrl ;					//!< [C^tF[X
		IVDMODCPUCTRL				cpuCtrl ;						//!< CPU C^tF[X
		IVDMODREGISTERCTRL			registerCtrl ;					//!< WX^C^tF[X
		VDMODMEMORYINFO				memoryInfo ;					//!< [
		vdmgr_wchar					memoryName[MEMORY_NAME_SIZE] ;	//!< [
		VDMODCPUCTRLINFO			cpuCtrlInfo ;					//!< CPU 
		VDMODREGISTERINFO			registerInfo[REGISTER_MAX] ;	//!< WX^
		vdmgr_wchar					registerName[REGISTER_NAME_SIZE] ;
																	//!< WX^
		vdmgr_uint					registerInfoCount ;				//!< WX^
		vdmgr_uint					registerNameLength ;			//!< WX^
		lua_State					*L ;							//!< Lua State

	} ;


// ------------------------------------------------------------------------------------------------ 
//   private                                                                                        

// -------------------------------------------------------------------------------- 
//!  [ݒ                                                               
//!                                                                                 
//!  @param [in,out] L  Lua State                                                   
//!  @return            ߂l                                                    
// -------------------------------------------------------------------------------- 
static int luafunc_setMemoryInfo( lua_State *L )
{
	VDMODLUA		*p=*(VDMODLUA **)lua_getextraspace( L ) ;
	const char		*p_string ;
	size_t			length ;
	lua_Integer		start, end, offset ;
	int				top ;


	do {
		top = lua_gettop( L ) ;
		if( (top!=3) && (top!=4) ) {
			break ;
		}

		start = lua_tointeger(L, 2) ;
		if( start > 0xffffffff ) {
			break ;
		}

		end = lua_tointeger(L, 3) ;
		if( (end>0xffffffff) || (start>end) ) {
			break ;
		}

		offset = 0 ;
		if( top == 4 ) {
			offset = lua_tointeger(L, 4) ;
			if( (offset<start) || (offset>end) ) {
				break ;
			}
		}

		p_string = lua_tolstring(L, 1, &length) ;
		if( !length ) {
			break ;
		}

		vdlib_utf8to16(&p_string[0], (vdmgr_int )length, &p->memoryName[0], MEMORY_NAME_SIZE) ;

		p->memoryInfo.start  = (vdmgr_u32 )start ;
		p->memoryInfo.end    = (vdmgr_u32 )end ;
		p->memoryInfo.offset = (vdmgr_u32 )offset ;

		return 0 ;
	} while( 0 ) ;

	vdlib_writeLog(VDLOG_ERROR VDMGR_WS("Lua : vdmod_setMemoryInfo(name, start, end [,offset])")) ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  CPU ݒ                                                                   
//!                                                                                 
//!  @param [in,out] L  Lua State                                                   
//!  @return            ߂l                                                    
// -------------------------------------------------------------------------------- 
static int luafunc_setCpuCtrlInfo( lua_State *L )
{
	VDMODLUA		*p=*(VDMODLUA **)lua_getextraspace( L ) ;
	lua_Integer		breakpointCount, startAddress, endAddress ;
	int				top ;


	do {
		top = lua_gettop( L ) ;
		if( top != 3 ) {
			break ;
		}

		breakpointCount = lua_tointeger(L, 1) ;

		startAddress = lua_tointeger(L, 2) ;
		if( startAddress > 0xffffffff ) {
			break ;
		}

		endAddress = lua_tointeger(L, 3) ;
		if( (endAddress>0xffffffff) || (startAddress>endAddress) ) {
			break ;
		}

		p->cpuCtrlInfo.breakpointCount = (vdmgr_uint )breakpointCount ;
		p->cpuCtrlInfo.startAddress    = (vdmgr_u32  )startAddress ;
		p->cpuCtrlInfo.endAddress      = (vdmgr_u32  )endAddress ;

		return 0 ;
	} while( 0 ) ;

	vdlib_writeLog(VDLOG_ERROR VDMGR_WS("Lua : luafunc_setCpuInfo(bkptCount, start, end)")) ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  WX^o^                                                                   
//!                                                                                 
//!  @param [in,out] L  Lua State                                                   
//!  @return            ߂l                                                    
// -------------------------------------------------------------------------------- 
static int luafunc_addRegister( lua_State *L )
{
	VDMODLUA			*p=*(VDMODLUA **)lua_getextraspace( L ) ;
	VDMODREGISTERINFO	*p_registerInfo ;
	lua_Integer			value ;
	const char			*p_name ;
	vdmgr_wchar			name[512] ;
	size_t				length ;
	vdmgr_int			result ;
	int					top ;


	if( p->registerInfoCount == REGISTER_MAX ) {
		return 0 ;
	}

	top = lua_gettop( L ) ;
	if( (top!=2) && (top!=4) ) {
		return 0 ;
	}

	p_name = lua_tolstring(L, 1, &length) ;
	result = vdlib_utf8to16(&p_name[0], (vdmgr_int )length, &name[0], VDMGR_COUNTOF( name )) ;

	if( (p->registerNameLength+result) >= REGISTER_NAME_SIZE ) {
		return 0 ;
	}

	value = lua_tointeger(L, 2) ;
	if( (value<=0) || (value>31) ) {
		return 0 ;
	}

	p_registerInfo         = &p->registerInfo[p->registerInfoCount] ;
	p_registerInfo->p_name = &p->registerName[p->registerNameLength] ;
	p_registerInfo->parent = 0xffff ;
	p_registerInfo->count  = (vdmgr_u8 )value ;

	if( top == 4 ) {

		// e
		value = lua_tointeger(L, 3) ;
		if( value >= p->registerInfoCount ) {
			return 1 ;
		}
		p_registerInfo->parent = (vdmgr_u16 )value ;

		// ItZbg
		value = lua_tointeger(L, 4) ;

		p_registerInfo->offset = (vdmgr_u8 )value ;
	}

	vdlib_wcscpy(&p_registerInfo->p_name[0], &name[0]) ;
	p->registerNameLength += (result + 1) ;

	p->registerInfoCount++ ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  NbNWFl[^~                                                       
//!                                                                                 
//!  @param [in,out] L  Lua State                                                   
//!  @return            ߂l                                                    
// -------------------------------------------------------------------------------- 
static int luafunc_stopClockGen( lua_State *L )
{
	static const vdmgr_int	reason[]={
								VDMODCLOCKGEN_REASON_ABORT,
								VDMODCLOCKGEN_REASON_CPUSTEP,
								VDMODCLOCKGEN_REASON_CPUBREAKPOINT,
								VDMODCLOCKGEN_REASON_CPUUNDEFINST
							} ;
	VDMODLUA				*p=*(VDMODLUA **)lua_getextraspace( L ) ;
	lua_Integer				index ;


	index = lua_tointeger(L, 1) ;
	if( index < VDMGR_COUNTOF( reason ) ) {
		vdmod_clockgen_stop(p->p_clockGen, reason[index]) ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  NbNCxgt                                                       
//!                                                                                 
//!  @param [in,out] L  Lua State                                                   
//!  @return            ߂l                                                    
// -------------------------------------------------------------------------------- 
static int luafunc_attachClockEvent( lua_State *L )
{
	VDMODLUA		*p=*(VDMODLUA **)lua_getextraspace( L ) ;
	lua_Integer		index ;


	index = lua_tointeger(L, 1) ;
	if( index < CLOCKEVENT_MAX ) {
		vdmod_clockgen_reattachClockEvent(p->p_clockGen, &p->clockEventInfo[index].clockEvent, (vdmgr_u32 )lua_tointeger(L, 2)) ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  NbNCxgO                                                       
//!                                                                                 
//!  @param [in,out] L  Lua State                                                   
//!  @return            ߂l                                                    
// -------------------------------------------------------------------------------- 
static int luafunc_detachClockEvent( lua_State *L )
{
	VDMODLUA		*p=*(VDMODLUA **)lua_getextraspace( L ) ;
	lua_Integer		index ;


	index = lua_tointeger(L, 1) ;
	if( index < CLOCKEVENT_MAX ) {
		vdmod_clockgen_detachClockEvent(p->p_clockGen, &p->clockEventInfo[index].clockEvent) ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  sǂݎ                                                                   
//!                                                                                 
//!  @param [in,out] L  Lua State                                                   
//!  @return            ߂l                                                    
// -------------------------------------------------------------------------------- 
static int luafunc_readPin( lua_State *L )
{
	VDMODLUA		*p=*(VDMODLUA **)lua_getextraspace( L ) ;
	lua_Integer		index ;


	index = lua_tointeger(L, 1) ;
	if( index >= PIN_MAX ) {
		return 0 ;
	}

	lua_pushinteger(L, p->pinInfo[index].pin.value) ;

	return 1 ;
}


// -------------------------------------------------------------------------------- 
//!  s                                                                   
//!                                                                                 
//!  @param [in,out] L  Lua State                                                   
//!  @return            ߂l                                                    
// -------------------------------------------------------------------------------- 
static int luafunc_writePin( lua_State *L )
{
	VDMODLUA		*p=*(VDMODLUA **)lua_getextraspace( L ) ;
	lua_Integer		index ;


	index = lua_tointeger(L, 1) ;
	if( index < VDMGR_COUNTOF( p->pinInfo ) ) {
		vdmod_pin_write(p->pinInfo[index].pin.p_output, (vdmgr_s32 )lua_tointeger(L, 2)) ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  AhXoXǂݎ                                                           
//!                                                                                 
//!  @param [in,out] L  Lua State                                                   
//!  @return            ߂l                                                    
// -------------------------------------------------------------------------------- 
static int luafunc_readAddressBus( lua_State *L )
{
	VDMODLUA	*p=*(VDMODLUA **)lua_getextraspace( L ) ;


	lua_pushinteger(L, vdmod_addressbus_read(p->p_addressBus, (vdmgr_u32 )lua_tointeger(L, 1))) ;

	return 1 ;
}


// -------------------------------------------------------------------------------- 
//!  AhXoX                                                           
//!                                                                                 
//!  @param [in,out] L  Lua State                                                   
//!  @return            ߂l                                                    
// -------------------------------------------------------------------------------- 
static int luafunc_writeAddressBus( lua_State *L )
{
	VDMODLUA		*p=*(VDMODLUA **)lua_getextraspace( L ) ;
	lua_Integer		address, value, size ;


	address = lua_tointeger(L, 1) ;
	value   = lua_tointeger(L, 2) ;
	size    = lua_tointeger(L, 3) ;

	switch( size ) {
		case 8:
		{
			vdmod_addressbus_write8(p->p_addressBus, (vdmgr_u32 )address, (vdmgr_u8 )value) ;
			break ;
		}

		case 16:
		{
			vdmod_addressbus_write16(p->p_addressBus, (vdmgr_u32 )address, (vdmgr_u16 )value) ;
			break ;
		}

		case 32:
		{
			vdmod_addressbus_write32(p->p_addressBus, (vdmgr_u32 )address, (vdmgr_u32 )value) ;
			break ;
		}

		default:
			break ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  O                                                                   
//!                                                                                 
//!  @param [in,out] L  Lua State                                                   
//!  @return            ߂l                                                    
// -------------------------------------------------------------------------------- 
static int luafunc_writeLog( lua_State *L )
{
	const char		*p_string ;
	size_t			length ;
	vdmgr_wchar		buff[1024] ;


	p_string = lua_tolstring(L, 1, &length) ;
	if( (p_string!=NULL) && length ) {
		vdlib_utf8to16(&p_string[0], (vdmgr_int )length, &buff[0], VDMGR_COUNTOF( buff )) ;
		vdlib_writeLog( &buff[0] ) ;
	}

	return 0 ;
}




// ------------------------------------------------------------------------------------------------ 
//   IVDMOD                                                                                         

// -------------------------------------------------------------------------------- 
//!                                                                             
//!                                                                                 
//!  @param [in,out] p_this  this |C^                                          
//!  @return                 Ȃ                                                   
// -------------------------------------------------------------------------------- 
static void VDMGR_METHODCALL module_release( IVDMOD *p_this )
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, vdmod, p_this) ;


	lua_close( p->L ) ;

	vdlib_free( p ) ;
}


// -------------------------------------------------------------------------------- 
//!  Rg[                                                                   
//!                                                                                 
//!  @param [in,out]  p_this   this |C^                                        
//!  @param [in]      control  Rg[ԍ                                     
//!  @param [in]      param    Pp[^                                       
//!  @param [in,out]  p_param  Qp[^                                       
//!  @retval          VDMGR_S_OK                                                    
//!  @retval          VDMGR_E_INVALIDPARAM                                          
//!  @retval          VDMGR_E_INVALIDCALL                                           
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL module_control(IVDMOD *p_this, vdmgr_u32 control, vdmgr_u32 param, void *p_param)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, vdmod, p_this) ;


	switch( control & 0xffff0000 ) {
		case VDCTRL_BUILDCOMPLETE:
		{
			lua_getglobal(p->L, "onBuildComplete") ;
			lua_pushinteger(p->L, vdmod_clockgen_getClockFreq( p->p_clockGen )) ;
			lua_pcall(p->L, 1, 0, 0) ;
			break ;
		}

		case VDCTRL_DESTROY:
		{
			lua_getglobal(p->L, "onDestroy") ;
			lua_pcall(p->L, 0, 0, 0) ;
			break ;
		}

		case VDCTRL_RESET:
		{
			lua_getglobal(p->L, "onReset") ;
			lua_pushinteger(p->L, param!=0) ;
			lua_pcall(p->L, 1, 0, 0) ;
			break ;
		}

		case VDCTRL_START:
		{
			lua_getglobal(p->L, "onStart") ;
			lua_pushinteger(p->L, param!=0) ;
			lua_pcall(p->L, 1, 0, 0) ;
			break ;
		}

		case VDCTRL_STOP:
		{
			lua_getglobal(p->L, "onStop") ;
			lua_pushinteger(p->L, param!=0) ;
			lua_pcall(p->L, 1, 0, 0) ;
			break ;
		}

		case VDCTRL_SETCLOCKGEN:
		{
			p->p_clockGen = (IVDMODCLOCKGEN *)p_param ;
			break ;
		}

		case VDCTRL_GETPIN:
		{
			if( (vdmgr_u16 )control >= PIN_MAX ) {
				return VDMGR_E_INVALIDPARAM ;
			}

			*(IVDMODPIN **)p_param = &p->pinInfo[(vdmgr_u16 )control].pin.input ;
			break ;
		}

		case VDCTRL_SETPIN:
		{
			if( (vdmgr_u16 )control >= PIN_MAX ) {
				return VDMGR_E_INVALIDPARAM ;
			}

			p->pinInfo[(vdmgr_u16 )control].pin.p_output = (IVDMODPIN *)p_param ;
			break ;
		}

		case VDCTRL_GETADDRESSBUS:   { *(IVDMODADDRESSBUS   **)p_param = &p->addressBus ;               break ; }
		case VDCTRL_SETADDRESSBUS:   { p->p_addressBus                 = (IVDMODADDRESSBUS *)p_param ;  break ; }
		case VDCTRL_GETMEMORYCTRL:   { *(IVDMODMEMORYCTRL   **)p_param = &p->memoryCtrl ;               break ; }
		case VDCTRL_GETCPUCTRL:      { *(IVDMODCPUCTRL      **)p_param = &p->cpuCtrl ;                  break ; }
		case VDCTRL_GETREGISTERCTRL: { *(IVDMODREGISTERCTRL **)p_param = &p->registerCtrl ;             break ; }

		default:
			return VDMGR_E_INVALIDCALL ;
	}

	return VDMGR_S_OK ;
}




// ------------------------------------------------------------------------------------------------ 
//   VDMODCLOCKEVENT                                                                                

// -------------------------------------------------------------------------------- 
//!  Cxg                                                                   
//!                                                                                 
//!  @param [in,out] p_this  this |C^                                          
//!  @return                 NbN                                             
// -------------------------------------------------------------------------------- 
static vdmgr_u32 VDMGR_METHODCALL clockevent_process( VDMODCLOCKEVENT *p_this )
{
	CLOCKEVENTINFO	*p_clockEventInfo=(CLOCKEVENTINFO *)p_this ;
	VDMODLUA		*p=p_clockEventInfo->p ;
	

	lua_getglobal(p->L, "onEventProcess") ;
	lua_pushinteger(p->L, (vdmgr_uint )(p_clockEventInfo-p->clockEventInfo)) ;
	lua_pcall(p->L, 1, 1, 0) ;

	return (vdmgr_u32 )lua_tointeger(p->L, -1) ;
}




// ------------------------------------------------------------------------------------------------ 
//   IVDMODPIN                                                                                      

// -------------------------------------------------------------------------------- 
//!                                                                         
//!                                                                                 
//!  @param [in,out] p_this  this |C^                                          
//!  @param [in]     value   l                                                     
//!  @return                 Ȃ                                                   
// -------------------------------------------------------------------------------- 
static void VDMGR_METHODCALL pin_write(IVDMODPIN *p_this, vdmgr_s32 value)
{
	PININFO		*p_pinInfo=(PININFO *)p_this ;
	VDMODLUA	*p=p_pinInfo->p ;


	lua_getglobal(p->L, "onWritePin") ;
	lua_pushinteger(p->L, (vdmgr_uint )(p_pinInfo-p->pinInfo)) ;
	lua_pushinteger(p->L, value) ;
	lua_pcall(p->L, 2, 0, 0) ;
}




// ------------------------------------------------------------------------------------------------ 
//   IVDMODADDRESSBUS                                                                               

// -------------------------------------------------------------------------------- 
//!  ǂݍ                                                                       
//!                                                                                 
//!  @param [in,out] p_this   this |C^                                         
//!  @param          address  AhX                                              
//!  @return                  l                                                    
// -------------------------------------------------------------------------------- 
static vdmgr_u32 VDMGR_METHODCALL addressbus_read(IVDMODADDRESSBUS *p_this, vdmgr_u32 address)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, addressBus, p_this) ;


	lua_getglobal(p->L, "onReadAddressBus") ;
	lua_pushinteger(p->L, address) ;
	lua_pcall(p->L, 1, 1, 0) ;

	return (vdmgr_u32 )lua_tointeger(p->L, -1) ;
}


// -------------------------------------------------------------------------------- 
//!  8bit                                                                   
//!                                                                                 
//!  @param [in,out] p_this   this |C^                                         
//!  @param [in]     address  AhX                                              
//!  @param [in]     value    l                                                    
//!  @return                  Ȃ                                                  
// -------------------------------------------------------------------------------- 
static void VDMGR_METHODCALL addressbus_write8(IVDMODADDRESSBUS *p_this, vdmgr_u32 address, vdmgr_u8 value)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, addressBus, p_this) ;


	lua_getglobal(p->L, "onWriteAddressBus") ;
	lua_pushinteger(p->L, address) ;
	lua_pushinteger(p->L, value) ;
	lua_pushinteger(p->L, 8) ;
	lua_pcall(p->L, 3, 0, 0) ;
}


// -------------------------------------------------------------------------------- 
//!  16bit                                                                  
//!                                                                                 
//!  @param [in,out] p_this   this |C^                                         
//!  @param [in]     address  AhX                                              
//!  @param [in]     value    l                                                    
//!  @return                  Ȃ                                                  
// -------------------------------------------------------------------------------- 
static void VDMGR_METHODCALL addressbus_write16(IVDMODADDRESSBUS *p_this, vdmgr_u32 address, vdmgr_u16 value)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, addressBus, p_this) ;


	lua_getglobal(p->L, "onWriteAddressBus") ;
	lua_pushinteger(p->L, address) ;
	lua_pushinteger(p->L, value) ;
	lua_pushinteger(p->L, 16) ;
	lua_pcall(p->L, 3, 0, 0) ;
}


// -------------------------------------------------------------------------------- 
//!  32bit                                                                  
//!                                                                                 
//!  @param [in,out] p_this   this |C^                                         
//!  @param [in]     address  AhX                                              
//!  @param [in]     value    l                                                    
//!  @return                  Ȃ                                                  
// -------------------------------------------------------------------------------- 
static void VDMGR_METHODCALL addressbus_write32(IVDMODADDRESSBUS *p_this, vdmgr_u32 address, vdmgr_u32 value)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, addressBus, p_this) ;


	lua_getglobal(p->L, "onWriteAddressBus") ;
	lua_pushinteger(p->L, address) ;
	lua_pushinteger(p->L, value) ;
	lua_pushinteger(p->L, 32) ;
	lua_pcall(p->L, 3, 0, 0) ;
}




// ------------------------------------------------------------------------------------------------ 
//   IVDMODMEMORYCTRL                                                                               

// -------------------------------------------------------------------------------- 
//!  [擾                                                             
//!                                                                                 
//!  @param [in,out] p_this         this |C^                                   
//!  @param [out]    pp_memoryInfo  [AhX̊i[AhX              
//!  @return                        [                                  
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL memoryctrl_getMemoryInfo(IVDMODMEMORYCTRL *p_this, VDMODMEMORYINFO **pp_memoryInfo)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, memoryCtrl, p_this) ;


	*pp_memoryInfo = &p->memoryInfo ;

	return 1 ;
}


// -------------------------------------------------------------------------------- 
//!  񂩂AhX擾                                                       
//!                                                                                 
//!  @param [in,out] p_this     this |C^                                       
//!  @param [in]     p_string                                                 
//!  @param [out]    p_index    [CfbNXi[AhX                    
//!  @param [out]    p_address  AhXi[AhX                                
//!  @retval         VDMGR_S_OK                                                     
//!  @retval         VDMGR_E_INVALIDPARAM                                           
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL memoryctrl_getAddress(IVDMODMEMORYCTRL *p_this, const vdmgr_wchar *p_string, vdmgr_uint *p_index, vdmgr_u32 *p_address)
{
	return VDMGR_E_INVALIDPARAM ;
}


// -------------------------------------------------------------------------------- 
//!  PoCgǂݍ                                                               
//!                                                                                 
//!  @param [in,out] p_this   this |C^                                         
//!  @param [in]     address  AhX                                              
//!  @return                  l                                                    
// -------------------------------------------------------------------------------- 
static vdmgr_u8 VDMGR_METHODCALL memoryctrl_read(IVDMODMEMORYCTRL *p_this, vdmgr_u32 address)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, memoryCtrl, p_this) ;


	lua_getglobal(p->L, "onReadAddressBus") ;
	lua_pushinteger(p->L, address) ;
	lua_pcall(p->L, 1, 1, 0) ;

	return (vdmgr_u8 )lua_tointeger(p->L, -1) ;
}


// -------------------------------------------------------------------------------- 
//!  PoCg                                                               
//!                                                                                 
//!  @param [in,out] p_this   this |C^                                         
//!  @param [in]     address  AhX                                              
//!  @param [in]     value    l                                                    
//!  @return         VDMGR_S_OK                                                     
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL memoryctrl_write(IVDMODMEMORYCTRL *p_this, vdmgr_u32 address, vdmgr_u8 value)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, memoryCtrl, p_this) ;


	lua_getglobal(p->L, "onWriteAddressBus") ;
	lua_pushinteger(p->L, address) ;
	lua_pushinteger(p->L, value) ;
	lua_pushinteger(p->L, 8) ;
	lua_pcall(p->L, 3, 0, 0) ;

	return VDMGR_S_OK ;
}




// ------------------------------------------------------------------------------------------------ 
//   IVDMODCPUCTRL                                                                                  

// -------------------------------------------------------------------------------- 
//!  CPU 擾                                                             
//!                                                                                 
//!  @param [in,out] p_this         this |C^                                   
//!  @param [out]    p_cpuCtrlInfo  CPU i[AhX                        
//!  @return                        Ȃ                                            
// -------------------------------------------------------------------------------- 
static void VDMGR_METHODCALL cpuctrl_getCpuCtrlInfo(IVDMODCPUCTRL *p_this, VDMODCPUCTRLINFO *p_cpuCtrlInfo)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, cpuCtrl, p_this) ;


	*p_cpuCtrlInfo = p->cpuCtrlInfo ;
}


// -------------------------------------------------------------------------------- 
//!  CXgNVAhX擾                                               
//!                                                                                 
//!  @param [in,out] p_this  this |C^                                          
//!  @return                 CXgNVAhX                             
// -------------------------------------------------------------------------------- 
static vdmgr_u32 VDMGR_METHODCALL cpuctrl_getInstAddress( IVDMODCPUCTRL *p_this )
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, cpuCtrl, p_this) ;


	lua_getglobal(p->L, "onGetInstAddress") ;
	lua_pcall(p->L, 0, 1, 0) ;

	return (vdmgr_u32 )lua_tointeger(p->L, -1) ;
}


// -------------------------------------------------------------------------------- 
//!  CXgNVAhXݒ                                               
//!                                                                                 
//!  @param [in,out] p_this   this |C^                                         
//!  @param [in]     address  AhX                                              
//!  @retval         VDMGR_S_OK                                                     
//!  @retval         VDMGR_E_INVALIDPARAM                                           
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL cpuctrl_setInstAddress(IVDMODCPUCTRL *p_this, vdmgr_u32 address)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, cpuCtrl, p_this) ;


	lua_getglobal(p->L, "onSetInstAddress") ;
	lua_pushinteger(p->L, address) ;
	lua_pcall(p->L, 1, 0, 0) ;

	return VDMGR_S_OK ;
}


// -------------------------------------------------------------------------------- 
//!  lCeBuAhX擾                                                       
//!                                                                                 
//!  @param [in,out] p_this   this |C^                                         
//!  @param [in]     address  AhX                                              
//!  @return                  AhX                                              
// -------------------------------------------------------------------------------- 
static vdmgr_u32 VDMGR_METHODCALL cpuctrl_getNativeAddress(IVDMODCPUCTRL *p_this, vdmgr_u32 address)
{
	return (vdmgr_u8 )address ;
}


// -------------------------------------------------------------------------------- 
//!  u[N|Cgݒ                                                         
//!                                                                                 
//!  @param [in,out] p_this   this |C^                                         
//!  @param [in]     address  AhX                                              
//!  @retval         VDMGR_S_OK                                                     
//!  @retval         VDMGR_E_INVALIDPARAM                                           
//!  @retval         VDMGR_E_INVALIDCALL                                            
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL cpuctrl_setBreakpoint(IVDMODCPUCTRL *p_this, vdmgr_u32 address)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, cpuCtrl, p_this) ;
	int			result ;


	lua_getglobal(p->L, "onSetBreakpoint") ;
	lua_pushinteger(p->L, address) ;

	result = 0 ;
	if( lua_pcall(p->L, 1, 1, 0) == LUA_OK ) {
		result = lua_toboolean(p->L, -1) ;
	}

	if( !result ) {
		return VDMGR_E_INVALIDCALL ;
	}

	return VDMGR_S_OK ;
}


// -------------------------------------------------------------------------------- 
//!  u[N|Cg                                                         
//!                                                                                 
//!  @param [in,out] p_this   this |C^                                         
//!  @param [in]     address  AhX                                              
//!  @retval         VDMGR_S_OK                                                     
//!  @retval         VDMGR_E_INVALIDPARAM                                           
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL cpuctrl_deleteBreakpoint(IVDMODCPUCTRL *p_this, vdmgr_u32 address)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, cpuCtrl, p_this) ;
	int			result ;


	lua_getglobal(p->L, "onDeleteBreakpoint") ;
	lua_pushinteger(p->L, address) ;

	result = 0 ;
	if( lua_pcall(p->L, 1, 1, 0) == LUA_OK ) {
		result = lua_toboolean(p->L, -1) ;
	}

	if( !result ) {
		return VDMGR_E_INVALIDCALL ;
	}

	return VDMGR_S_OK ;
}


// -------------------------------------------------------------------------------- 
//!  XebvC                                                                   
//!                                                                                 
//!  @param [in,out] p_this  this |C^                                          
//!  @retval         VDMGR_S_OK                                                     
//!  @retval         VDMGR_E_INVALIDCALL                                            
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL cpuctrl_stepIn( IVDMODCPUCTRL *p_this )
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, cpuCtrl, p_this) ;


	lua_getglobal(p->L, "onStepIn") ;
	lua_pcall(p->L, 0, 0, 0) ;

	return VDMGR_S_OK ;
}


// -------------------------------------------------------------------------------- 
//!  XebvI[o[                                                               
//!                                                                                 
//!  @param [in,out] p_this  this |C^                                          
//!  @retval         VDMGR_S_OK                                                     
//!  @retval         VDMGR_E_INVALIDCALL                                            
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL cpuctrl_stepOver( IVDMODCPUCTRL *p_this )
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, cpuCtrl, p_this) ;


	lua_getglobal(p->L, "onStepOver") ;
	lua_pcall(p->L, 0, 0, 0) ;

	return VDMGR_S_OK ;
}


// -------------------------------------------------------------------------------- 
//!  wAhX̖ߕ쐬                                                 
//!                                                                                 
//!  @param [in,out] p_this       this |C^                                     
//!  @param [in]     address      AhX                                          
//!  @param [out]    p_string     i[AhX                                
//!  @param [in]     p_debugInfo  fobO                                      
//!  @return                      R[hTCY                                      
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL cpuctrl_createInstString(IVDMODCPUCTRL *p_this, vdmgr_u32 address, vdmgr_wchar *p_string, IVDMODDEBUGINFO *p_debugInfo)
{
	VDMODLUA		*p=VDMGR_THISPTR(VDMODLUA, cpuCtrl, p_this) ;
	const char		*p_instString ;
	lua_Integer		size ;


	if( address > p->cpuCtrlInfo.endAddress ) {
		return 0 ;
	}

	lua_getglobal(p->L, "onCreateInstString") ;
	lua_pushinteger(p->L, address) ;

	p_instString = NULL ;
	size         = 1 ;
	if( lua_pcall(p->L, 1, 2, 0) == LUA_OK ) {
		p_instString = lua_tostring(p->L,  -1) ;
		size         = lua_tointeger(p->L, -2) ;
	}

	if( p_instString!= NULL ) {
		vdlib_utf8to16(&p_instString[0], -1, &p_string[0], 1024) ;
	} else {
		vdlib_wcssprintf(&p_string[0], VDMGR_WS("%04X  ??"), address) ;
	}

	size += (size == 0) ;

	return (vdmgr_int )size ;
}




// ------------------------------------------------------------------------------------------------ 
//   IVDMODREGISTERCTRL                                                                             

// -------------------------------------------------------------------------------- 
//!  WX^AhX擾                                                     
//!                                                                                 
//!  @param [in,out] p_this           this |C^                                 
//!  @param [out]    pp_registerInfo  WX^AhX̊i[AhX            
//!  @return                          WX^                                
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL registerctrl_getRegisterInfo(IVDMODREGISTERCTRL *p_this, VDMODREGISTERINFO **pp_registerInfo)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, registerCtrl, p_this) ;


	*pp_registerInfo = p->registerInfo ;

	return p->registerInfoCount ;
}


// -------------------------------------------------------------------------------- 
//!  WX^̒l擾                                                             
//!                                                                                 
//!  @param [in,out] p_this   this |C^                                         
//!  @param [in]     index    WX^CfbNX                                  
//!  @param [out]    p_value  li[AhX                                        
//!  @retval         VDMGR_S_OK                                                     
//!  @retval         VDMGR_E_INVALIDPARAM                                           
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL registerctrl_getValue(IVDMODREGISTERCTRL *p_this, vdmgr_u16 index, vdmgr_u32 *p_value)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, registerCtrl, p_this) ;


	lua_getglobal(p->L, "onGetRegister") ;
	lua_pushinteger(p->L, (vdmgr_uint )index) ;
	lua_pcall(p->L, 1, 1, 0) ;

	*p_value = (vdmgr_u32 )lua_tointeger(p->L, -1) ;

	return VDMGR_S_OK ;
}


// -------------------------------------------------------------------------------- 
//!  WX^̒lݒ                                                             
//!                                                                                 
//!  @param [in,out] p_this  this |C^                                          
//!  @param [in]     index   WX^CfbNX                                   
//!  @param [in]     value   l                                                     
//!  @retval         VDMGR_S_OK                                                     
//!  @retval         VDMGR_E_INVALIDPARAM                                           
// -------------------------------------------------------------------------------- 
static vdmgr_int VDMGR_METHODCALL registerctrl_setValue(IVDMODREGISTERCTRL *p_this, vdmgr_u16 index, vdmgr_u32 value)
{
	VDMODLUA	*p=VDMGR_THISPTR(VDMODLUA, registerCtrl, p_this) ;


	lua_getglobal(p->L, "onSetRegister") ;
	lua_pushinteger(p->L, index) ;
	lua_pushinteger(p->L, value) ;
	lua_pcall(p->L, 2, 0, 0) ;

	return VDMGR_S_OK ;
}




// -------------------------------------------------------------------------------- 
//!   W[擾                                                          
//!                                                                                 
//!   @param [out] p_vdmodInfo  W[i[AhX                          
//!   @return                   Ȃ                                                
// -------------------------------------------------------------------------------- 
void vdmod_lua_getModuleInfo( VDMODINFO *p_vdmodInfo )
{
	static const VDMODITEM	item[]={
								{ VDMGR_WS("pin0"),         VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     0  },
								{ VDMGR_WS("pin1"),         VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     1  },
								{ VDMGR_WS("pin2"),         VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     2  },
								{ VDMGR_WS("pin3"),         VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     3  },
								{ VDMGR_WS("pin4"),         VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     4  },
								{ VDMGR_WS("pin5"),         VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     5  },
								{ VDMGR_WS("pin6"),         VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     6  },
								{ VDMGR_WS("pin7"),         VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     7  },
								{ VDMGR_WS("pin8"),         VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     8  },
								{ VDMGR_WS("pin9"),         VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     9  },
								{ VDMGR_WS("pin10"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     10 },
								{ VDMGR_WS("pin11"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     11 },
								{ VDMGR_WS("pin12"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     12 },
								{ VDMGR_WS("pin13"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     13 },
								{ VDMGR_WS("pin14"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     14 },
								{ VDMGR_WS("pin15"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     15 },
								{ VDMGR_WS("pin16"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     16 },
								{ VDMGR_WS("pin17"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     17 },
								{ VDMGR_WS("pin18"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     18 },
								{ VDMGR_WS("pin19"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     19 },
								{ VDMGR_WS("pin20"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     20 },
								{ VDMGR_WS("pin21"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     21 },
								{ VDMGR_WS("pin22"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     22 },
								{ VDMGR_WS("pin23"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     23 },
								{ VDMGR_WS("pin24"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     24 },
								{ VDMGR_WS("pin25"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     25 },
								{ VDMGR_WS("pin26"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     26 },
								{ VDMGR_WS("pin27"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     27 },
								{ VDMGR_WS("pin28"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     28 },
								{ VDMGR_WS("pin29"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     29 },
								{ VDMGR_WS("pin30"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     30 },
								{ VDMGR_WS("pin31"),        VDMODITEM_TYPE_PIN,          VDMODPIN_SPEC_DIGITAL_INOUT,                                                     31 },
								{ VDMGR_WS("memoryctrl"),   VDMODITEM_TYPE_MEMORYCTRL,   0,                                                                               0  },
								{ VDMGR_WS("addressbus"),   VDMODITEM_TYPE_ADDRESSBUS,   VDMODADDRESSBUS_SPEC_GETSET|VDMODADDRESSBUS_SPEC_BUS|VDMODADDRESSBUS_SPEC_WRITE, 0  },
								{ VDMGR_WS("cpuctrl"),      VDMODITEM_TYPE_CPUCTRL,      0,                                                                               0  },
								{ VDMGR_WS("registerctrl"), VDMODITEM_TYPE_REGISTERCTRL, 0,                                                                               0  }
							} ;


	p_vdmodInfo->p_name          = VDMGR_WS("lua") ;
	p_vdmodInfo->version         = 0x00000001 ;
	p_vdmodInfo->type            = VDMODINFO_TYPE_GENERAL2 ;
	p_vdmodInfo->p_item          = &item[0] ;
	p_vdmodInfo->itemCount       = VDMGR_COUNTOF( item ) ;
	p_vdmodInfo->create.general2 = vdmod_lua_create ;
}


// -------------------------------------------------------------------------------- 
//!  W[쐬                                                               
//!                                                                                 
//!  @param [in] p_vdmgr     zfoCXǗ                                       
//!  @param [in] p_param     Pp[^                                         
//!  @param [in] param       Qp[^                                         
//!  @retval     VDMGR_NULL  ُI                                               
//!  @retval     LȊO    W[C^tF[X                               
// -------------------------------------------------------------------------------- 
IVDMOD *vdmod_lua_create(VDMGR *p_vdmgr, const vdmgr_wchar *p_param, vdmgr_u32 param)
{
	VDMODLUA		*p ;
	vdmgr_char		fileName[VDMGR_MAXPATH] ;
	vdmgr_wchar		buff[1024] ;
	vdmgr_int		result, index ;

	// [m
	p = (VDMODLUA *)vdlib_zalloc( sizeof( VDMODLUA ) ) ;
	if( p == VDMGR_NULL ) {
		return VDMGR_NULL ;
	}
	p->p_vdmgr                      = p_vdmgr ;
	p->vdmod.release                = module_release ;
	p->vdmod.control                = module_control ;
	p->addressBus.read              =addressbus_read ;
	p->addressBus.write8            =addressbus_write8 ;
	p->addressBus.write16           =addressbus_write16 ;
	p->addressBus.write32           =addressbus_write32 ;
	p->p_addressBus                 = vdlib_addressbus_get() ;
	p->memoryCtrl.getMemoryInfo     = memoryctrl_getMemoryInfo ;
	p->memoryCtrl.getAddress        = memoryctrl_getAddress ;
	p->memoryCtrl.read              = memoryctrl_read ;
	p->memoryCtrl.write             = memoryctrl_write ;
	p->cpuCtrl.getCpuCtrlInfo       = cpuctrl_getCpuCtrlInfo ;
	p->cpuCtrl.getInstAddress       = cpuctrl_getInstAddress ;
	p->cpuCtrl.setInstAddress       = cpuctrl_setInstAddress ;
	p->cpuCtrl.getNativeAddress     = cpuctrl_getNativeAddress ;
	p->cpuCtrl.setBreakpoint        = cpuctrl_setBreakpoint ;
	p->cpuCtrl.deleteBreakpoint     = cpuctrl_deleteBreakpoint ;
	p->cpuCtrl.stepIn               = cpuctrl_stepIn ;
	p->cpuCtrl.stepOver             = cpuctrl_stepOver ;
	p->cpuCtrl.createInstString     = cpuctrl_createInstString ;
	p->registerCtrl.getRegisterInfo = registerctrl_getRegisterInfo ;
	p->registerCtrl.getValue        = registerctrl_getValue ;
	p->registerCtrl.setValue        = registerctrl_setValue ;
	p->memoryInfo.p_name            = &p->memoryName[0] ;
	p->memoryInfo.offset            = 0 ;
	p->memoryInfo.start             = 0 ;
	p->memoryInfo.end               = 15 ;
	p->cpuCtrlInfo.breakpointCount  = 1 ;
	p->cpuCtrlInfo.startAddress     = 0 ;
	p->cpuCtrlInfo.endAddress       = 15 ;

	vdlib_wcscpy(&p->memoryName[0], VDMGR_WS("lua - unknown")) ;

	for(index=0; index<CLOCKEVENT_MAX; index++) {
		if( !index ) {
			vdmod_clockevent_init(&p->clockEventInfo[index].clockEvent, VDMODCLOCKEVENT_PRIO_IO,                           clockevent_process) ;
		} else {
			vdmod_clockevent_init(&p->clockEventInfo[index].clockEvent, (vdmgr_s16 )(VDMODCLOCKEVENT_PRIO_NORMAL+index-1), clockevent_process) ;
		}
		p->clockEventInfo[index].p = p ;
	}

	for(index=0; index<PIN_MAX; index++) {
		vdlib_pin_init(&p->pinInfo[index].pin, VDLIBPIN_TYPE_INVERT, pin_write, 0) ;
		p->pinInfo[index].p = p ;
	}

	// Lua
	p->L = luaL_newstate() ;
	if( p->L == NULL ) {
		vdlib_free( p ) ;
		return VDMGR_NULL ;
	}

	*(VDMODLUA **)lua_getextraspace( p->L ) = p ;

	lua_register(p->L, "vdmod_setMemoryInfo",    &luafunc_setMemoryInfo) ;
	lua_register(p->L, "vdmod_setCpuCtrlInfo",   &luafunc_setCpuCtrlInfo) ;
	lua_register(p->L, "vdmod_addRegister",      &luafunc_addRegister) ;
	lua_register(p->L, "vdmod_stopClockGen",     &luafunc_stopClockGen) ;
	lua_register(p->L, "vdmod_attachClockEvent", &luafunc_attachClockEvent) ;
	lua_register(p->L, "vdmod_detachClockEvent", &luafunc_detachClockEvent) ;
	lua_register(p->L, "vdmod_readPin",          &luafunc_readPin) ;
	lua_register(p->L, "vdmod_writePin",         &luafunc_writePin) ;
	lua_register(p->L, "vdmod_readAddressBus",   &luafunc_readAddressBus) ;
	lua_register(p->L, "vdmod_writeAddressBus",  &luafunc_writeAddressBus) ;
	lua_register(p->L, "vdmod_writeLog",         &luafunc_writeLog) ;

	luaL_openlibs( p->L ) ;

	vdlib_wcstombs(&p_param[0], -1, &fileName[0], sizeof( fileName )) ;

	result = luaL_dofile(p->L, &fileName[0]) ;
	if( result ) {
		vdlib_utf8to16(lua_tostring(p->L, -1), -1, &buff[0], VDMGR_COUNTOF( buff )) ;
		vdlib_writeLog(VDLOG_ERROR VDMGR_WS("Lua : %s"), &buff[0]) ;

		lua_close( p->L ) ;
		vdlib_free( p ) ;

		return VDMGR_NULL ;
	}

	return &p->vdmod ;
}
